/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

use crate::config::ProjectConfig;
use common::Diagnostic;
use dependency_analyzer::get_reachable_ir;
use fnv::FnvHashMap;
use graphql_syntax::ExecutableDefinition;
use graphql_text_printer::print_executable_definition_ast;
use intern::string_key::StringKey;
use md5::{Digest, Md5};
use relay_transforms::DependencyMap;
use schema::SDLSchema;

use super::ProjectAsts;

pub struct BuildIRResult {
    pub ir: Vec<graphql_ir::ExecutableDefinition>,
    pub source_hashes: SourceHashes,
}

/// Map fragments and queries definition names to the md5 of they printed source
pub struct SourceHashes(FnvHashMap<StringKey, String>);

impl SourceHashes {
    pub fn from_definitions(definitions: &[ExecutableDefinition]) -> Self {
        let mut source_hashes = FnvHashMap::default();
        for ast in definitions {
            if let Some(name) = ast.name() {
                source_hashes.insert(name, md5(&print_executable_definition_ast(ast)));
            }
        }
        Self(source_hashes)
    }

    pub fn get(&self, k: &StringKey) -> Option<&String> {
        self.0.get(k)
    }
}

pub fn build_ir(
    _project_config: &ProjectConfig,
    implicit_dependencies: &DependencyMap,
    project_asts: ProjectAsts,
    schema: &SDLSchema,
    is_incremental_build: bool,
) -> Result<BuildIRResult, Vec<Diagnostic>> {
    let asts = project_asts.definitions;
    let source_hashes = SourceHashes::from_definitions(&asts);
    let ir = graphql_ir::build_ir_in_relay_mode(schema, &asts)?;
    if is_incremental_build {
        let affected_ir = get_reachable_ir(
            ir,
            project_asts.base_definition_names,
            project_asts.changed_names,
            implicit_dependencies,
            schema,
        );
        Ok(BuildIRResult {
            ir: affected_ir,
            source_hashes,
        })
    } else {
        Ok(BuildIRResult { ir, source_hashes })
    }
}

fn md5(data: &str) -> String {
    let mut md5 = Md5::new();
    md5.input(data);
    hex::encode(md5.result())
}
